"""
Constants and functions shared between modules.
"""

# This is hardcoded in doc8 as a valid option so be wary that changing this
# may break doc8 linting.
# See https://github.com/PyCQA/doc8/pull/34.
SUBSTITUTION_OPTION_NAME = "substitutions"

CONTENT_SUBSTITUTION_OPTION_NAME = "content-substitutions"
PATH_SUBSTITUTION_OPTION_NAME = "path-substitutions"
NO_SUBSTITUTION_OPTION_NAME = "nosubstitutions"
NO_CONTENT_SUBSTITUTION_OPTION_NAME = "nocontent-substitutions"
NO_PATH_SUBSTITUTION_OPTION_NAME = "nopath-substitutions"
